#ifdef WIN32
# include "connection_win32.cxx"
#else

#include "connection.h"
#include <sys/types.h>
#include <sys/wait.h>
#include <signal.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdio.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>

static int set_nonblock_flag (int desc, int value)
{
  int oldflags = fcntl (desc, F_GETFL, 0);
  /* If reading the flags failed, return error indication now. */
  if (oldflags == -1)
    return -1;
  /* Set just the flag we want to set. */
  if (value != 0)
    oldflags |= O_NONBLOCK;
  else
    oldflags &= ~O_NONBLOCK;
  /* Store modified flag word in the descriptor. */
  return fcntl (desc, F_SETFL, oldflags);
}


static bool init_sockaddr (struct sockaddr_in *name,
			   const char *hostname,
			   uint16_t port)
{
  struct hostent *hostinfo;
  
  name->sin_family = AF_INET;
  name->sin_port = htons (port);
  hostinfo = gethostbyname (hostname);
  if (hostinfo == NULL) {
    fprintf (stderr, "Unknown host %s.\n", hostname);
    return 0;
  }
  name->sin_addr = *(struct in_addr *) hostinfo->h_addr;
  return 1;
}

Connection::Connection()
{
  sock = fdin = fdout = -1;
}


Connection::~Connection()
{
  close();
}


bool Connection::open( char*s, int p )
{
  sockaddr_in name;

  sock = socket(PF_INET, SOCK_STREAM, 0);
  if (sock < 0) {
    fprintf(stderr, "Cannot create socket !\n");
    return 0;
  }

  if (!init_sockaddr (&name, s, p)) return 0;

  if (0 > connect(sock, (sockaddr*) &name, sizeof(name))) {
    perror("Connection::open");
    close();
    return 0;
  }

/*  if (fcntl(sock, F_SETFL, O_NONBLOCK) == -1) {
    fprintf(stderr, "WARNING : Could not execute fcntl on socket");
    //    return 0;
  }*/
  set_nonblock_flag(sock, 0);

  fdin = sock;
  fdout = sock;
  
  return 1;
}

void Connection::close()
{
  if (sock > 0)
    ::close(sock);
  sock = 0;
}

int Connection::read(void* buf, int sz)
{
  return ::read(fdin, buf, sz);
}

int Connection::write(const void* buf, int sz)
{
  return ::write(fdout, buf, sz);
}

#endif
