class GameBrowser;
#include "game_browser.h"
#include <FL/fl_draw.H>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "vizigo.h"

static Stone* stn;

GameBrowser::GameBrowser(int x, int y, int w, int h, char* l) : 
  Fl_Browser(x,y,w,h,l)

{
  for (int i=0; i<MAX_IGSGAMES; i++) games[i] = 0;
  nbgames = 0;
  my_list.b = this;
  widget.b = this;
  list(&my_list);
  parent()->remove(widget); // Yuck ! This is a problem because FLTK's 
                            // automatic add to groups feature !!
  type(type()|VERTICAL_ALWAYS);
}

GameBrowser::~GameBrowser()
{
  nbgames = 0;
}

inline int result_adjust(const void* pa, const void* pb)
{
  IgsGame *a = cur_igs->games[*(int*)pa], *b = cur_igs->games[*(int*)pb];

  return (a->result && (cur_time-a->time_finished > 60*TIME_FACTOR)? 32000:0) - 
	 (b->result && (cur_time-b->time_finished > 60*TIME_FACTOR)? 32000:0);
}

static int wpc(const void* pa, const void* pb)
{
  IgsGame *a = cur_igs->games[*(int*)pa], *b = cur_igs->games[*(int*)pb];
  if (a->whitename == 0) return -1 + result_adjust(pa, pb);
  if (b->whitename == 0) return +1 + result_adjust(pa, pb);
  return strcmp(a->whitename, b->whitename) + result_adjust(pa, pb);
}

static int wrc(const void* pa, const void* pb)
{
  IgsGame *a = cur_igs->games[*(int*)pa], *b = cur_igs->games[*(int*)pb];
  char *sa = a->whiterank, *sb = b->whiterank;
  return rank_compare(sa, sb) + result_adjust(pa, pb);
}
static int bpc(const void* pa, const void* pb)
{
  IgsGame *a = cur_igs->games[*(int*)pa], *b = cur_igs->games[*(int*)pb];
  if (a->blackname == 0) return -1 + result_adjust(pa, pb);
  if (b->blackname == 0) return +1 + result_adjust(pa, pb);
  return strcmp(a->blackname, b->blackname) + result_adjust(pa, pb);
}
static int brc(const void* pa, const void* pb)
{
  IgsGame *a = cur_igs->games[*(int*)pa], *b = cur_igs->games[*(int*)pb];
  char *sa = a->blackrank, *sb = b->blackrank;
  return rank_compare(sa, sb) + result_adjust(pa, pb);
}

static int nbmc(const void* pa, const void* pb)
{
  IgsGame *a = cur_igs->games[*(int*)pa], *b = cur_igs->games[*(int*)pb];
  return b->curpos-a->curpos + result_adjust(pa, pb);
}

static int komic(const void* pa, const void* pb)
{
  IgsGame *a = cur_igs->games[*(int*)pa], *b = cur_igs->games[*(int*)pb];
  return int(10*(a->komi-b->komi)) + result_adjust(pa, pb);
}

static int sizec(const void* pa, const void* pb)
{
  IgsGame *a = cur_igs->games[*(int*)pa], *b = cur_igs->games[*(int*)pb];
  return a->size-b->size + result_adjust(pa, pb);
}

static int handic(const void* pa, const void* pb)
{
  IgsGame *a = cur_igs->games[*(int*)pa], *b = cur_igs->games[*(int*)pb];
  return b->handi-a->handi + result_adjust(pa, pb);
}

void GameBrowser::sort()
{
  gigs(this);
  int i, j;
  for (i=1, j=1; i<MAX_IGSGAMES; i++)
    if (cur_igs->games[i]) games[j++] = i;
  nbgames = j-1;

  int (*compar)(const void *, const void *) = 0;
  if (cur_igs->UI->wp->value()) compar = wpc;
  else if (cur_igs->UI->wr->value()) compar = wrc;
  else if (cur_igs->UI->bp->value()) compar = bpc;
  else if (cur_igs->UI->br->value()) compar = brc;
  else if (cur_igs->UI->nbm->value()) compar = nbmc;
  else if (cur_igs->UI->komi->value()) compar = komic;
  else if (cur_igs->UI->size->value()) compar = sizec;
  else if (cur_igs->UI->handi->value()) compar = handic;
  if (compar) qsort(games+1, nbgames, sizeof(int), compar);
  else {
    nbgames = games[nbgames]+1;
    for (i=1; i<nbgames; i++) games[i] = i;
  }
  layout();
}

int GameBrowser::List::children(const Fl_Menu_*, const int* indexes, int level)
{
  if (level) return 0;
  if (b->nbgames == 0) return 0;
  return b->nbgames-1;
}

Fl_Widget* GameBrowser::List::child(const Fl_Menu_*, const int* indexes, int level)
{
  int n = *indexes;
  if (level || n<0 || n>=b->nbgames-1) return 0;
  b->widget.n = n+1;
  gigs(b);
  IgsGame* g = cur_igs->games[b->games[n+1]];
  if (g && g->observed)
    b->widget.set_flag(FL_VALUE);
  else
    b->widget.clear_flag(FL_VALUE);
  return &b->widget;
}

#define W 650-20
#define H 13
GameBrowser::Widget::Widget() : 
  Fl_Widget(0, 0, W, H, 0)
{
}

GameBrowser::Widget::~Widget()
{
}

void GameBrowser::Widget::draw()
{
  char s[128];
  int x = this->x(), y = this->y(), w = this->w(), h = this->h();
  gigs(b);
  int n = b->games[this->n];
  IgsGame* g = cur_igs->games[n];
#ifdef WIN32
  //fl_font(b->text_font(), b->text_size()-2);
  fl_font(FIXED_FONT, FIXED_FONTSIZE);
#else
  fl_font(b->text_font(), b->text_size());
#endif
  if (g && g->whitename) {
    char f[20];
    f[3] = 0;
    f[0] = g->flags&IGSGAME_T? 'T':' ';
    f[1] = g->flags&IGSGAME_F? 'F':' ';
    f[2] = g->flags&IGSGAME_I? 'I':' ';
    if (g->result) {
      if ((g->result&IGSGAME_ADJ) == IGSGAME_ADJ) strcpy(f, "Adj.");
      else {
	char r[3];
	if (g->result&IGSGAME_WWINS) strcpy(r, "W+");
	else strcpy(r, "B+");
	if (g->result&IGSGAME_TIME) sprintf(f, "%sTime", r);
	else if (g->result&IGSGAME_RESIGN) sprintf(f, "%sRes.", r);
	else sprintf(f, "%s%.1f", r, g->result_score);
      }
    }
    fl_color(b->text_color());
    sprintf(s, "[%3d ]", n);
    fl_draw(s, x, y, w, h, FL_ALIGN_LEFT);
//    fl_font(b->text_font(), b->text_size());
#ifdef WIN3
# define W1 "13"
# define W2 "5"
# define W3 "15"
# define W4 "5"
# define W5 "10"
# define W6 "6"
# define W7 "7"
# define W8 "5"
# define W9 "5"
#else
# define W1 "11"
# define W2 "4"
# define W3 "12"
# define W4 "4"
# define W5 "7"
# define W6 "5"
# define W7 "6"
# define W8 "4"
# define W9 "4"
#endif
    sprintf(s, "%"W1"s [ %"W2"s ]     %"W3"s [ %"W4"s ] %"W5"s  %"W6"d %"W7".1f %"W8"d %"W9"d", 
	    g->whitename, g->whiterank, g->blackname, g->blackrank, f, 
	    g->curpos, g->komi, g->size, g->handi);
    fl_draw(s, x+6*12, y, w-6*12, h, FL_ALIGN_LEFT);
    if (!stn) stn = small_stone;
    stn->draw(x+5*12, y, 1+n%DRAWSTONE_NUMWHITE);
    stn->draw(x+20*12, y, 0);
  } else {
    fl_color(b->text_color());
//    fl_font(b->text_font(), b->text_size());
    sprintf(s, "[%3d ]", n);
    fl_draw(s, x, y, w, h, FL_ALIGN_LEFT);
  }
}

int GameBrowser::Widget::handle(int event) {
  //  printf("GameBrowser::Widget::handle\n");
  int n;
  IgsGame* g;
  switch (event) {
  case FL_RELEASE:
    gigs(b);
    n = b->games[this->n];
    cur_igs->observe(n);
    g = cur_igs->games[n];
    b->set_focus();
    b->select(n, g->observed);
    b->redraw();
  case FL_PUSH:
  case FL_DRAG:
    return 1;
    
  case FL_KEYBOARD:
    switch (Fl::event_key()) {
    case ' ':
      gigs(b);
      n = b->games[this->n];
      cur_igs->observe(n);
      g = cur_igs->games[n];
      //      b->goto_number(n);
      b->set_focus();
      b->redraw();
      return 1;
    default:
      return 0;
    }

  default:
    return 0;
  }
}

void GameBrowser::click()
{
  gigs(this);
  int n = value()+1;
  if (n>0 && n<MAX_IGSGAMES && (n = games[n])) {
    if (Fl::event_button() == 1)
      cur_igs->observe(n);
    else {
      int x = Fl::event_x()-this->x();
      char* name = 0;
      if (x>=40 && x<40+120)
	name = cur_igs->games[n]->whitename;
      else if (x>=215 && x<215+125)
	name = cur_igs->games[n]->blackname;
      if (name) {
	char s[128];
	sprintf(s, "stats %s", name);
	cur_igs->send(s);
      }
    }
  }
}
