#include "board.h"
#include "vizigo.h"
#include <FL/fl_draw.H>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include "stone.h"

#define stn small_stone

#define XSHIFT 30
#define WS 60

#define LINK_COL 128, 128, 128
#define SEL_LINK_COL 255, 192, 192

static void cb(Fl_Widget* o, void* p)
{
  MovesBrowser* b = (MovesBrowser*) o;

  //  printf("callback\n");
  if (!Fl::event_button()) return;
  int n = b->value()-1;
  if (b->board && b->board->mode) {
    b->board->go(n+1);
    return;
  }
  int x = (Fl::event_x()+b->xposition()-XSHIFT-b->x())/WS;
  //  printf("%d\n", x);
  if (n>=0) {
    Story* s = b->moves[n+1];
    while (s) {
      if (s->indent == x) break;
      s = s->brother;
    }
    if (s) {
      while (s->past) {
	Story* ns = s;
	s = s->past;
	n--;
	int i;
	for (i=0; i<s->nbvariants && s->variants[i]!=ns; i++);
	if (i<s->nbvariants) s->current = i;
      }
      n = 0;
      while (s) {
	b->sel_moves[n++] = s;
	s = s->step();
      }
      while (n<b->nbmoves)
	b->sel_moves[n++] = 0;
      b->redraw();
    }
  }
  b->board->go(b->value());
}

MovesBrowser::MovesBrowser(int x, int y, int w, int h, char* l) : 
  Fl_Browser(x,y,w,h,l)

{
  max_indent = 0;
  nbmoves = 0;
  maxmoves = 16;
  moves = (Story**) malloc(sizeof(Story*)*maxmoves);
  sel_moves = (Story**) malloc(sizeof(Story*)*maxmoves);
  board = 0;
  my_list.b = this;
  widget.b = this;
  list(&my_list);
  parent()->remove(widget); // Yuck ! This is a problem because FLTK's 
                            // automatic add to groups feature !!
  type(type()|VERTICAL);
  widget.resize(0, 0, w-20, 20);
  callback(cb);
}

MovesBrowser::~MovesBrowser()
{
  free(moves);
  free(sel_moves);
}

int MovesBrowser::List::children(const Fl_Menu_*, const int* indexes, int level)
{
  if (level) return 0;
  return b->nbmoves+1;
}

#define W 600-20
#define H 20
Fl_Widget* MovesBrowser::List::child(const Fl_Menu_*, const int* indexes, int level)
{
  if (level) return 0;
  if (*indexes < 0 || *indexes >= b->nbmoves+1) return 0;
/*  if (b->board && (*indexes)-1 == b->board->pos)
    b->widget.set_flag(FL_VALUE);
  else
    b->widget.clear_flag(FL_VALUE);*/
  b->widget.n = (*indexes);
  b->widget.resize(0, 0, XSHIFT+WS*(b->max_indent+1), H);
  return &b->widget;
}

MovesBrowser::Widget::Widget() : Fl_Widget(0, 0, W, H, 0)
{
}

MovesBrowser::Widget::~Widget()
{
}

inline void my_line(int x1, int y1, int x2, int y2)
{
  fl_line_style(FL_CAP_ROUND, 2);
  fl_line(x1, y1, x2, y2);
  fl_line_style(0);
/*  fl_line(x1, y1+1, x2, y2+1);
  fl_line(x1, y1-1, x2, y2-1);
  fl_line(x1+1, y1, x2+1, y2);
  fl_line(x1-1, y1, x2-1, y2);*/
}

void MovesBrowser::Widget::draw()
{
  if (n<1) return;
  int x = this->x(), y = this->y(), w = this->w(), h = this->h();
  fl_clip(x, y, w, h);
#ifdef WIN32
  fl_font(b->text_font(), b->text_size()-2);
#else
  fl_font(b->text_font(), b->text_size());
#endif
  fl_color(b->text_color());
  char p[128];
  sprintf(p, "%d", n);
  fl_draw(p, x+4, y, w, h, FL_ALIGN_LEFT);
  Story* s = b->moves[n];
  while (s) {
    fl_color(fl_rgb(LINK_COL));
    //fl_color(s == b->sel_moves[n]? fl_rgb(SEL_LINK_COL) : fl_rgb(LINK_COL));
    if (s->past && s != b->sel_moves[n]) {
      //my_line(x+XSHIFT+s->indent*WS+12/2, y+h/2, x+XSHIFT+s->past->indent*WS+12/2, y+h/2-h);
      my_line(x+XSHIFT+s->past->indent*WS+12/2, y+h/2-h, x+XSHIFT+s->past->indent*WS+12/2, y);
      my_line(x+XSHIFT+s->indent*WS+12/2, y, x+XSHIFT+s->past->indent*WS+12/2, y);
      my_line(x+XSHIFT+s->indent*WS+12/2, y, x+XSHIFT+s->indent*WS+12/2, y+h/2);
    }
/*    for (int i=0; i<s->nbvariants; i++) {
      fl_color(s->variants[i] == b->sel_moves[n+1]? fl_rgb(SEL_LINK_COL) : fl_rgb(LINK_COL));
            my_line(x+XSHIFT+s->indent*WS+12/2, y+h/2, x+XSHIFT+s->variants[i]->indent*WS+12/2, y+h/2+h);
    }*/
    if (s->nbvariants) {
      fl_color(s == b->sel_moves[n]? fl_rgb(SEL_LINK_COL) : fl_rgb(LINK_COL));
      my_line(x+XSHIFT+s->indent*WS+12/2, y+h/2, x+XSHIFT+s->indent*WS+12/2, y+h/2+h);
    }
    s = s->brother;
  }

  s = b->sel_moves[n];
  if (s && s->past) {
    /* draw selected link last to erase unselected ones */
    fl_color(fl_rgb(SEL_LINK_COL));
    my_line(x+XSHIFT+s->past->indent*WS+12/2, y+h/2-h, x+XSHIFT+s->past->indent*WS+12/2, y);
    my_line(x+XSHIFT+s->indent*WS+12/2, y, x+XSHIFT+s->past->indent*WS+12/2, y);
    my_line(x+XSHIFT+s->indent*WS+12/2, y, x+XSHIFT+s->indent*WS+12/2, y+h/2);
  }
  
  s = b->moves[n];
  while (s) {
    stn->draw(x+XSHIFT+s->indent*WS+(12-stn->get_size())/2, y+(h-stn->get_size())/2, ~n&1);
    if (s->x >= 0) {
      char c = 'a'+s->x;
      if (c>='i') c++;
      sprintf(p, "%c%d", c, b->board->size-s->y);
    } else {
      if (s->y == -1)
	strcpy(p, "pass");
      else
	sprintf(p, "ha. %d", -s->y);
    }
    fl_color(fl_rgb(255, 255, 128));
    fl_draw(p, x+XSHIFT+s->indent*WS+14, y, WS-14, h, FL_ALIGN_LEFT);
    s = s->brother;
  }

  fl_pop_clip();
}

int MovesBrowser::Widget::handle(int event) {
  switch(event) {
/*  case FL_RELEASE:
    printf("RELEASE !!\n");
    return 1;*/
  default:
    return 0;
  }
}


void MovesBrowser::graph(int time, Story* s)
{
  if (time>=maxmoves) {
    maxmoves *= 2;
    moves = (Story**) realloc(moves, sizeof(Story*)*maxmoves);
    sel_moves = (Story**) realloc(sel_moves, sizeof(Story*)*maxmoves);
  }
  if (time>nbmoves) {
    nbmoves = time;
    moves[time] = 0;
  }
  if (0 && !moves[time])
    s->indent = 0;
  else {
    if (s->past) {
      int i = s->past->indent;
      i = max_indent;
      Story* ss = moves[time];
      while (ss) {
	if (ss->indent == i) break;
	ss = ss->brother;
      }
      while (ss) {
	i++;
	ss = moves[time];
	while (ss) {
	  if (ss->indent == i) break;
	  ss = ss->brother;
	}
      }
      s->indent = i;
    } else
      s->indent = 0;
  }
  max_indent = MAX(max_indent, s->indent);
  s->brother = moves[time];
  moves[time] = s;
  int mean = 0;
  for (int i=0; i<s->nbvariants; i++) {
    graph(time+1, s->variants[i]);
    mean += moves[time+1]->indent;
  }
  if (s->nbvariants) {
    s->indent = mean / s->nbvariants;
  }
}

void MovesBrowser::make_graph()
{
  Story* s = board->s;
  nbmoves = 0;
  max_indent = 0;
  if (s) {
    while (s->past) s = s->past;
    graph(0, s);
    int i=0;
    while (s) {
      sel_moves[i++] = s;
      s = s->step();
    }
    while (i<nbmoves)
      sel_moves[i++] = 0;
  }
  layout();
  redraw();
}


int MovesBrowser::handle(int event)
{
  switch(event) {
  case FL_FOCUS:
    return 0;
  default:
    return Fl_Browser::handle(event);
  }
}


int Fl_Tabs_NF::handle(int event)
{
  switch(event) {
  case FL_FOCUS:
    return 0;
  default:
    return Fl_Tabs::handle(event);
  }
}

int Fl_Group_NF::handle(int event)
{
  switch(event) {
  case FL_FOCUS:
    return 0;
  default:
    return Fl_Group::handle(event);
  }
}
