class ObserverBrowser;
#include "observer_browser.h"
#include "defs.h"
#include <FL/fl_draw.H>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>

ObserverBrowser::ObserverBrowser(int x, int y, int w, int h, char* l) : 
  Fl_Browser(x,y,w,h,l)

{
  observers = 0;
  nbobservers = 0;
  maxobservers = 0;
  my_list.b = this;
  widget.b = this;
  list(&my_list);
  parent()->remove(widget); // Yuck ! This is a problem because FLTK's 
                            // automatic add to groups feature !!
  type((type()|VERTICAL_ALWAYS)&(~HORIZONTAL));
  widget.resize(0, 0, w-20, 13);
}

ObserverBrowser::~ObserverBrowser()
{
  clear();
}

int ObserverBrowser::List::children(const Fl_Menu_*, const int* indexes, int level)
{
  if (level) return 0;
  return b->nbobservers;
}

Fl_Widget* ObserverBrowser::List::child(const Fl_Menu_*, const int* indexes, int level)
{
  if (level || *indexes < 0 || *indexes >= b->nbobservers) return 0;
  b->widget.name = b->observers[*indexes];
  return &b->widget;
}

#define W 600-20
#define H 13
ObserverBrowser::Widget::Widget() : Fl_Widget(0, 0, W, H, 0)
{
}

ObserverBrowser::Widget::~Widget()
{
}

void ObserverBrowser::Widget::draw()
{
  int x = this->x(), y = this->y(), w = this->w(), h = this->h();
  fl_color(b->text_color());
#ifdef WIN32
//  fl_font(b->text_font(), b->text_size()-2);
  fl_font(FIXED_FONT, FIXED_FONTSIZE);
#else
  fl_font(b->text_font(), b->text_size());
#endif
  fl_draw(name, x, y, w, h, FL_ALIGN_LEFT);
}

int ObserverBrowser::Widget::handle(int event) {
  return 0;
}

void ObserverBrowser::clear()
{
  deselect();
  if (observers) {
    for (int i=0; i<nbobservers; i++)
      free(observers[i]);
    free(observers);
    maxobservers = 0;
    nbobservers = 0;
    observers = 0;
  }
  layout();
  redraw();
}

void ObserverBrowser::add(const char* name)
{
  if (nbobservers >= maxobservers) {
    maxobservers *= 2;
    if (observers) 
      observers = (char**) realloc(observers, sizeof(int)*maxobservers);
    else {
      maxobservers = 16;
      observers = (char**) malloc(sizeof(int)*maxobservers);
    }
  }
  observers[nbobservers++] = strdup(name);
  layout();
  redraw();
}
