class PlayerBrowser;
#include "player_browser.h"
#include <FL/fl_draw.H>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include "stone.h"
#include "igs.h"

PlayerBrowser::PlayerBrowser(int x, int y, int w, int h, char* l) : 
  Fl_Browser(x,y,w,h,l)

{
  cur_player = 0;
  players = 0;
  nbplayers = 0;
  my_list.b = this;
  widget.b = this;
  list(&my_list);
  parent()->remove(widget); // Yuck ! This is a problem because FLTK's 
                            // automatic add to groups feature !!
  type(type()|VERTICAL_ALWAYS);
  widget.resize(0, 0, w-20, 13);
}

PlayerBrowser::~PlayerBrowser()
{
  nbplayers = 0;
}

#ifdef WIN32x
int strcasecmp(const char* a, const char* b)
{
  while(*a && *b) {
    int diff = tolower(*a++) - tolower(*b++);
    if (diff) return diff;
  }
  return *a - *b;
}
#endif

int cmpa_player(const void* pa, const void* pb)
{
  IgsPlayer *a = *(IgsPlayer**)pa, *b = *(IgsPlayer**)pb;
  int d = strcasecmp(a->name, b->name);
  if (d == 0) d = strcmp(a->name, b->name);
  return d;
}

int cmpr_player(const void* pa, const void* pb)
{
  IgsPlayer *a = *(IgsPlayer**)pa, *b = *(IgsPlayer**)pb;
  int d = rank_compare(a->rank, b->rank);
  if (d == 0) d = (b->playing? 0:int(b->flags)&3) - (a->playing? 0:int(a->flags)&3);
  if (d == 0) d = cmpa_player(pa, pb);
  return d;
}

int cmps_player(const void* pa, const void* pb)
{
  IgsPlayer *a = *(IgsPlayer**)pa, *b = *(IgsPlayer**)pb;
  int d = (b->playing? 0:int(b->flags)&3) - (a->playing? 0:int(a->flags)&3);
  if (d == 0) return rank_compare(a->rank, b->rank);
  else return d;
}

void PlayerBrowser::sort()
{
  gigs(this);
  if (topline() < nbplayers) {
    layout();
    top_player = players[topline()];
    //    DBFprintf("top player '%s'\n", top_player->name);
  } else
    top_player = 0;
  if (players) free(players);
  players = (IgsPlayer**) malloc(sizeof(IgsPlayer*)*cur_igs->nbplayers);
  nbplayers = cur_igs->nbplayers;
  memcpy(players, cur_igs->players, sizeof(IgsPlayer*)*nbplayers);

  int (*compar)(const void *, const void *) = cmpa_player;
  if (cur_igs->UI->pr->value()) compar = cmpr_player;
  else if (cur_igs->UI->ps->value()) compar = cmps_player;
  qsort(players, nbplayers, sizeof(IgsPlayer*), compar);
  if (cur_player) {
/*    IgsPlayer* k = (IgsPlayer*) malloc (sizeof(IgsPlayer));
    k->name = cur_player->name;
    k->rank = cur_player->rank;
    IgsPlayer** p = (IgsPlayer**) bsearch(&k, players, nbplayers, 
					  sizeof(IgsPlayer*), compar);
    if (p) select(p-players, 1);
    else   {
      cur_player = 0;
      cur_igs->update_player();
    }
    free(k);*/
    int i;
    for (i=0; i<nbplayers; i++) if (cur_player == players[i]) break;
    if (i<nbplayers) {
      select(i, 1);
      //      center(i);
    } else {
      cur_player = 0;
      cur_igs->update_player();
    }
  }
  if (top_player) {
    int i;
    for (i=0; i<nbplayers; i++) if (top_player == players[i]) break;
    if (i<nbplayers)
      topline(i);
  }
  layout();
}

void PlayerBrowser::click()
{
  int n = value();
  cur_player = players[n];
  gigs(this);
  cur_igs->update_player();
  if (!Fl::event_clicks() && !cur_player->defaults) {
    char s[128];
    sprintf(s, "stats %s", cur_player->name);
    cur_igs->send(s);
    DBGprintf("ASK STATS !!\n");
  }
}

void PlayerBrowser::select_player(IgsPlayer* p)
{
  gigs(this);
  cur_player = p;
  int i;
  for (i=0; i<nbplayers; i++) if (cur_player == players[i]) break;
  if (i<nbplayers) {
    select(i, 1);
    //    center(i);
  click();
  } else {
    cur_player = 0;
    deselect();
    cur_igs->update_player();
  }
  cur_igs->UI->tabs->value(cur_igs->UI->players_group);
}

void PlayerBrowser::damage_player(IgsPlayer* p)
{
/*  gigs(this);
  int i;
  for (i=0; i<nbplayers; i++) if (p == players[i]) break;
  if (i<nbplayers)
    damage_item(i);*/
  redraw();
}

int PlayerBrowser::List::children(const Fl_Menu_*, const int* indexes, int level)
{
  if (level) return 0;
  return b->nbplayers;
}

Fl_Widget* PlayerBrowser::List::child(const Fl_Menu_*, const int* indexes, int level)
{
  if (level || *indexes < 0 || *indexes >= b->nbplayers) return 0;
  b->widget.p = b->players[*indexes];
  return &b->widget;
}

#define W 600-20
#define H 13
PlayerBrowser::Widget::Widget() : Fl_Widget(0, 0, W, H, 0)
{
}

PlayerBrowser::Widget::~Widget()
{
}

void PlayerBrowser::Widget::draw()
{
  int x = this->x(), y = this->y(), w = this->w(), h = this->h();
  gigs(b);
  char* f = "?";
  switch (p->flags&3) {
  case IPF_CLOSED:
    f = "X";
    break;
  case IPF_OPEN:
    f= "O";
    break;
  case IPF_LOOKING:
    f = "!";
    break;
  default:
    break;
  }
  if (p->playing) f = "X";
  char s[128];
  fl_color(b->text_color());
#ifdef WIN32
//  fl_font(b->text_font(), b->text_size()-2);
//# define W1 "15"
//# define W2 "5"
//  fl_font(b->text_font(),12 , fl_encoding, 7);
  fl_font(FIXED_FONT, FIXED_FONTSIZE);
# define W1 "13"
# define W2 "4"
#else
  fl_font(b->text_font(), b->text_size());
# define W1 "13"
# define W2 "4"
#endif
  sprintf(s, "%"W1"s [ %"W2"s ] %s", p->name, p->rank, f);
  fl_draw(s, x, y, w, h, FL_ALIGN_LEFT);
}

int PlayerBrowser::Widget::handle(int event) {
  return 0;
}
