//
//                   **** ViZiGO (C) 2000 Vincent Penne ****
// This software if distributed under the GPL license (See the file LICENSE
// that must be included with this source)
//
/** 
 *  @file player_browser.h
 *
 * PlayerBrowser class definition
 *
 */

#ifndef PLAYERBROWSER_H
#define PLAYERBROWSER_H

#include <FL/Fl_Browser.H>
#include "igsplayer.h"

/**
 * class PlayerBrowser
 */
class PlayerBrowser : public Fl_Browser {
 private:

  /**
   * The widget used to draw one player
   */
  class Widget : public Fl_Widget {
  public:
    Widget();
    ~Widget();
    void draw();
    int handle(int event);
    IgsPlayer* p;
    PlayerBrowser* b;
  };
  friend class PlayerBrowser::Widget;
  Widget widget;         ///< one widget for all players

  /**
   * The list used by the Fl_Browser
   */
  class List : public Fl_List {
  public:
    virtual int children(const Fl_Menu_*, const int* indexes, int level);
    virtual Fl_Widget* child(const Fl_Menu_*, const int* indexes, int level);
    PlayerBrowser* b;
  };
  friend class PlayerBrowser::List;
  List my_list;          ///< instance of the list

 public:

  IgsPlayer* cur_player; ///< selected player
  IgsPlayer* top_player; ///< player on top of browser player
  IgsPlayer** players;   ///< sorted list of players
  int nbplayers;         ///< number of players

  PlayerBrowser(int x, int y, int w, int h, char* l = 0);
  ~PlayerBrowser();
  
  void sort();           ///< sort the player list
  void click();          ///< click on player callback
  void select_player(IgsPlayer* p); ///< select a player
  void damage_player(IgsPlayer* p); ///< redraw a player
  
};

#endif // ifdef PLAYERBROWSER_H
