#include "defs.h"
#include "board.h"
#include "igsgame.h"
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>

// is it a symbol ?
static bool is(char c)
{
  switch(c) {
  case '(': case ')': case '[': case ']': case ';': case 10 : case 13:
    //  case ')': case ')': case ')': case ')': case ')': case ')': case ')':
    return 1;

  default:
    return 0;
  };
}

static char *p, *buffer, *end;

static char* getword()
{
  static char s[10*1024];
  while (p<end && isspace(*p)) p++;
  if (is(*p)) {
    *s = *p;
    s[1] = 0;
    p++;
    DBGprintf("<<%s>>", s); fflush(stdout);
    return s;
  }
  char *ps = s;
  while (p<end && !is(*p)) *ps++ = *p++;
  *ps++ = 0;
  DBGprintf("<%s>", s); fflush(stdout);
  return s;
}

static char* getbracket(bool a=0)
{
  bool escape = 0;
  static char s[10*1024];
  int num = 1;
  *s = 0;
  char* w;
  if (!a) {
    w = getword();
    if (*w != '[') {DBGprintf("no bracket !! %s\n", w); return "";}
  }
  do {
    //    w = getword();
    if (!escape) {
      escape = *p == '\\';
      if (*p == ']') num--;
      if (*p == '[') num++;
    } else
      escape = 0;
    if (num>0) { char w[2]; w[0]=*p; w[1] = 0; strcat(s, w); }
    p++;
  } while (num>0 && p<end);
  //  DBGprintf("[%s]", s); fflush(stdout);
  return s;
}

void Board::load_sgf(char* name)
{
  if (name) {
    memset(inib, 0, 19*19);
    clear();
    
    FILE* fp = fopen(name, "rb");
    if (!fp) return;

    int sz = 100*1024;
    buffer = (char*) malloc(sz);
    end = buffer+fread(buffer, 1, sz, fp);
    while (end == buffer + sz) {
      buffer = (char *) realloc(buffer, 2*sz);
      end = buffer+sz+fread(buffer+sz, 1, sz, fp);
      sz *= 2;
    }
    fclose(fp);
    p = buffer;
    if (strcmp(getword(), "(")) {
      free(buffer);
      return;
    }
  }
  int mx = 32000, my = 32000;
  for (;;) {
    char *s = getword();
    if (!strcmp(s, ";")) {
      if (mx < 32000) {
	add_move(mx, my, 0);
	mx = 32000;
      }
    }
    else if (!strcmp(s, "GM")) {
      //      DBGprintf("GM --> %c\n", *p);
      getbracket();
    } 
    else if (!strcmp(s, "SZ")) {
      //      DBGprintf("GM --> %c\n", *p);
      set_size(atoi(getbracket()));
      DBGprintf("Size = %d\n", size);
    } 
    else if (!strcmp(s, "W") ||!strcmp(s, "B")) {
      cur_player = *s == 'W'? WHITE:BLACK;
      char* c = getbracket();
      if (!strcmp(c, "tt") || strlen(c) < 2) 
	mx = my = -1;
      else {
	mx = tolower(c[0])-'a';
	my = tolower(c[1])-'a';
      }
      add_move(mx, my, 0);
      mx = 32000;
    } 
    else if (*s == '(') {
      if (mx < 32000) {
	add_move(mx, my, 0);
	mx = 32000;
      }

      int keep_pos = pos;
      load_sgf();
      go(keep_pos, 0);
    }
    else if (*s == ')' || p>=end) break;
    else if (!strcmp(s, "KM")) {
      s = getbracket();
      score[0] = 0;
      score[1] = atof(s);
    }
    else if (*s == 'A' && (s[1] == 'W' || s[1] == 'B' || s[1] == 'E')) {
      //      add_move(-1, -1);
      char color = s[1] == 'W'? WHITE: (s[1] == 'B'? BLACK:EMPTY);
      for (;;) {
	char* oldp = p;
	s = getword(); 
	if (*s != '[') {
	  p = oldp;
	  break;
	}
	s = getbracket(1);
	int i = tolower(s[0])-'a', j=tolower(s[1])-'a';
	b[i][j] = b[i][j] == color;
	if (this->s)
	  this->s->b[i][j] = color;
	else
	  inib[i][j] = color;
      }
    } 
    else if (!strcmp(s, "HA")) {
      cur_player = BLACK;
      int h = atoi(getbracket());
      mx = -1;
      my = -h;
      add_move(mx, my, 0);
      mx = 32000;
#define GET_STRING(id, string)    } else if (!strcmp(s, id)) { \
                       strncpy(string, getbracket(), sizeof(string)-1)
    GET_STRING("PW", whitename);
    GET_STRING("WR", whiterank);
    GET_STRING("PB", blackname);
    GET_STRING("BR", blackrank);
    } 
    else if (*s=='C') {
      if (this->s) {
	if (this->s->commentary)
	  free(this->s->commentary);
	this->s->commentary = strdup(getbracket());
      } else 
	UI->output->add(getbracket());
    } 
    else if (*s == '[') {
      getbracket(1);
    }
    else getbracket();

    if (p>=end) break;
  }

  if (mx < 32000) {
    add_move(mx, my, 0);
    mx = 32000;
  }

  if (name) {
    /* load other root variants */
/*    while (!strcmp(getword(), "(")) {
      go(0);
      load_sgf();
      
    } */

    free(buffer);
    UI->moves_browser->make_graph();
    go(0);
    while (s->nbvariants) { // select the primary story
      s->current = s->nbvariants-1;
      s = s->step();
    }
    go(32000);
  }
}

void Board::save_sgf(char* name)
{
  static FILE* fp;
  int oldpos;
  if (name) {
    fp = fopen(name, "wb");
    if (!fp) return;

    oldpos = pos;
    go(1);

    fprintf(fp, 
	    "(;GM[1]FF[3]\n"
	    "SZ[%d]KM[%1.1f]\n"
	    "PW[%s]\n"
	    "WR[%s]\n"
	    "PB[%s]\n"
	    "BR[%s]\n"
	    "SY[%s];\n",
	    size,
	    owner? owner->komi:s->score[1], 
	    *whitename? whitename:"unknown", 
	    *whiterank? whiterank:"NR", 
	    *blackname? blackname:"unknown", 
	    *blackrank? blackrank:"NR",
	    "ViZiGO "VERSION_STRING);
  }
  for (;;) {
    Story* os = s;
    if (s->commentary) 
      fprintf(fp, "C[%s]", s->commentary);
    if (s->x >= 0)
      fprintf(fp, "%s[%c%c];\n", s->player == BLACK? "B":"W", char(s->x + 'a'),
	      char(s->y + 'a'));
    else {
      if (s->y == -1)
	fprintf(fp, "%s[tt];\n", s->player == BLACK? "B":"W");
      else
	fprintf(fp, "HA[%d];\n", -s->y);
    }
    if (s->nbvariants == 0) break;
    int i = 0;
    for (; i<s->nbvariants-1; i++) {
      s = os->variants[i];
      fprintf(fp, "(;");
      save_sgf();
      fprintf(fp, ")");
    }
    s = os->variants[i];
  }
  if (name) {
    fputc(')', fp);
    fclose(fp);
    go(oldpos);
  }
}
