#include "stone.h"
#include "Fl_RGBM_Image.h"
#include <stdlib.h>
#include <stdio.h>
#include <math.h>

uchar cols[256*3];

static char* unique()
{
  static int n;
  static char s[128];
  sprintf(s, "stone%d", n++);
  return s;
}

// from cgoban
static int  yinYangX(int size, int y)  {
  float  center;
  float  fy = y + 0.5;

  if (y < size / 2)  {
    center = (float)size * 0.25;
    return((int)((float)size * 0.5 +
		 sqrt(center * center -
		      (fy - center) * (fy - center)) +
		 0.5));
  } else  {
    center = (float)size * 0.75;
    return((int)((float)size * 0.5 -
		 sqrt((float)(size * size) * 0.0625 -
		      (fy - center) * (fy - center)) +
		 0.5));
  }
}

void Stone::draw(int x, int y, char c)
{
  init_image();
  img[c]->draw(x, y, size, size, 0, 0); 
}


void Stone::set_size(int s)
{
  if (size==s) return;
  size = s;
  if (img[0]) {
    for (int i=0; i<DRAWSTONE_NUMWHITE+1; i++) {
      img[i]->remove();
      img[i] = 0;
    }
    free(pixels[0]);
    free(pixels[1]);
  }
}

void Stone::init_image()
{
  if (img[0] == 0) {
    pixels[0] = (uint*)malloc(sizeof(int)*(size*size+2));
    pixels[1] = (uint*)malloc(sizeof(int)*(size*size+2)*DRAWSTONE_NUMWHITE);
    pixels[2] = (uint*)malloc(sizeof(int)*(size*size+2));
    { for (int i=0; i<256; i++) {
      cols[i] = i; 
      cols[i+256] = i;
      cols[i+512] = i;
    }
    }
    drawStone_newPics(cols, pixels[0], pixels[1], size);
    pixels[2][0] = pixels[0][0];
    pixels[2][1] = pixels[0][1];
    for (int i=0; i<size; i++)
      for (int j=0; j<size; j++)
	pixels[2][2+i+j*size] = pixels[i<yinYangX(size, j)? 0:1][2+i+j*size];
    if (type == STONE_GRAYED) {
      for (int i=0; i<size; i++)
	for (int j=0; j<size; j++) {
	  if ((i^j)&1) {
	    pixels[0][2+i+j*size] = 0xff000000;
	    for (int k=0; k<DRAWSTONE_NUMWHITE; k++)
	      pixels[1][2+i+j*size+k*(size*size+2)] = 0xff000000;
	  }
      }
    } else if (type == STONE_SHADOW) {
      for (int i=0; i<size; i++)
	for (int j=0; j<size; j++) {
	  if ((i^j)&1) {
	    pixels[0][2+i+j*size] = 0xff000000;
	    for (int k=0; k<DRAWSTONE_NUMWHITE; k++)
	      pixels[1][2+i+j*size+k*(size*size+2)] = 0xff000000;
	  } else if (!(pixels[0][2+i+j*size]&0xff000000)) {
	    pixels[0][2+i+j*size] = 0x000000;
	    for (int k=0; k<DRAWSTONE_NUMWHITE; k++)
	      pixels[1][2+i+j*size+k*(size*size+2)] = 0x000000;
	  }
      }
    }
    img[0] = Fl_RGBM_Image::get(unique(), (uchar*)pixels[0]);
    { for (int i=0; i<DRAWSTONE_NUMWHITE; i++)
      img[i+1] = Fl_RGBM_Image::get(unique(), (uchar*)(pixels[1]+(size*size+2)*i));
    }
    img[DRAWSTONE_NUMWHITE+1] = Fl_RGBM_Image::get(unique(), (uchar*)pixels[2]);
  }
}


Stone::~Stone()
{
  if (img[0]) {
    for (int i=0; i<DRAWSTONE_NUMWHITE+2; i++) {
      img[i]->remove();
      img[i] = 0;
    }
    free(pixels[0]);
    free(pixels[1]);
    free(pixels[2]);
  }
}
